###########################################################################
#  Vintel - Visual Intel Chat Analyzer                                    #
#  Copyright (C) 2014-15 Sebastian Meyer (sparrow.242.de+eve@gmail.com )  #
#                                                                         #
#  This program is free software: you can redistribute it and/or modify   #
#  it under the terms of the GNU General Public License as published by   #
#  the Free Software Foundation, either version 3 of the License, or      #
#  (at your option) any later version.                                    #
#                                                                         #
#  This program is distributed in the hope that it will be useful,        #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#  GNU General Public License for more details.                           #
#                                                                         #
#                                                                         #
#  You should have received a copy of the GNU General Public License      #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
###########################################################################

from PyQt4.QtCore import QThread
from PyQt4.QtCore import SIGNAL

from Queue import Queue
import time

from vi import evegate
from vi.cache.cache import Cache

class AvatarFindThread(QThread):

    def __init__(self):
        QThread.__init__(self)
        self.q = Queue()

    def add_chatentry(self, chatentry, clear_cache=False):
        if clear_cache:
            cache = Cache()
            cache.remove_avatar(chatentry.message.user)
        self.q.put(chatentry)

    def run(self):
        cache = Cache()
        last_call = 0
        wait = 300 # time between 2 requests in ms
        while True:
            chatentry = self.q.get()
            charname = chatentry.message.user
            avatar = cache.get_avatar(charname)
            if not avatar:
                diff_last_call = time.time() - last_call
                if diff_last_call < wait:
                    time.sleep((wait - diff_last_call) /1000.0)
                avatar = evegate.get_avatar_for_player(charname)
                last_call = time.time()
                if avatar:
                    cache.put_avatar(charname, avatar)
            if avatar:
                self.emit(SIGNAL("avatar_update"), chatentry, avatar)
            """
            # deactivated instand retry to load the avatar to prevent any
            # spamming. If it fails it fails. That's destiny ;)
            else:
                self.q.put(chatentry)
            """


class PlayerFindThread(QThread):
    
    def __init__(self):
        QThread.__init__(self)
        self.q = Queue()
        
    def add_chatentry(self, chatentry):
        self.q.put(chatentry)
        
    def run(self):
        cache = Cache()
        while True:
            chatentry = self.q.get()
            text = chatentry.message.utext
            parts = text.split("  ")  # split@double-space (not in name, right?)
            