###########################################################################
#  Vintel - Visual Intel Chat Analyzer                                    #
#  Copyright (C) 2014-15 Sebastian Meyer (sparrow.242.de+eve@gmail.com )  #
#                                                                         #
#  This program is free software: you can redistribute it and/or modify   #
#  it under the terms of the GNU General Public License as published by   #
#  the Free Software Foundation, either version 3 of the License, or      #
#  (at your option) any later version.                                    #
#                                                                         #
#  This program is distributed in the hope that it will be useful,        #
#  but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#  GNU General Public License for more details.                           #
#                                                                         #
#                                                                         #
#  You should have received a copy of the GNU General Public License      #
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
###########################################################################

import datetime
import os
import sys
import time
import webbrowser

from PyQt4 import Qt, QtGui, uic, QtCore
from PyQt4.QtGui import QImage, QPixmap
from PyQt4.QtWebKit import QWebPage
from PyQt4.QtCore import QPoint

from vi.ui.systemtray import TrayContextMenu
from vi.ui.threads import AvatarFindThread
import vi.version
from vi import chatparser, dotlan, filewatcher, koschecker
from vi import sound, drachenjaeger
from vi.cache.cache import Cache

from vi.resources import resource_path

VERSION = vi.version.VERSION


class MainWindow(QtGui.QMainWindow):

    def __init__(self, path_to_logs, trayicon):
        """ systems = list of system-objects creted by dotlan.py
        """
        QtGui.QMainWindow.__init__(self)
        uic.loadUi(resource_path('vi/ui/MainWindow.ui'), self)
        self.setWindowTitle("Vintel " + VERSION)
        self.setWindowIcon(QtGui.QIcon(
            resource_path("vi/ui/res/logo_small.png")))
        self.path_to_logs = path_to_logs
        self.trayicon = trayicon
        self.trayicon.activated.connect(self.systray_activated)
        self.dotlan = dotlan.Map("Providence", "Providence_test.svg")
        self.init_map_position = None  # we read this after first rendering
        self.set_map_content(self.dotlan.svg)
        self.systems = self.dotlan.systems
        self.chatentries = []
        self.avatar_find_thread = AvatarFindThread()
        self.connect(self.avatar_find_thread, QtCore.SIGNAL("avatar_update"),
                      self.update_avatar_on_chatentry)
        self.avatar_find_thread.start()
        self.clipboard = QtGui.QApplication.clipboard()
        self.clipboard.clear(mode=self.clipboard.Clipboard)
        self.old_clipboard_content = (0, "")
        self.connect(self.clipboard, Qt.SIGNAL("changed(QClipboard::Mode)"), 
                     self.clipboard_changed)
        self.connect(self.zoomin, Qt.SIGNAL("clicked()"), self.zoomMapIn)
        self.connect(self.zoomout, Qt.SIGNAL("clicked()"), self.zoomMapOut)
        self.connect(self.chat_large, Qt.SIGNAL("clicked()"), self.chat_larger)
        self.connect(self.chat_small, Qt.SIGNAL("clicked()"), self.chat_smaller)
        self.connect(self.actionInfo, Qt.SIGNAL("triggered()"), self.show_info)
        self.connect(self.actionShow_Chat_Avatars, Qt.SIGNAL("triggered()"), 
                     self.change_show_avatars)
        self.connect(self.actionAlways_on_top, Qt.SIGNAL("triggered()"),
                     self.change_always_on_top)
        self.connect(self.jumpBridgesButton, Qt.SIGNAL("clicked()"), 
                     self.change_jumpbridge_view)
        self.connect(self.choose_chatrooms_button, Qt.SIGNAL("triggered()"),
                     self.show_chatroom_chooser)
        self.connect(self.action_show_chat, Qt.SIGNAL("triggered()"),
                     self.change_chat_visibility)
        self.opacity_group = QtGui.QActionGroup(self.menu)
        for i in (100, 80, 60, 40, 20):
            action = QtGui.QAction("Opacity {0}%".format(i), None, 
                                   checkable=True)
            if i == 100:
                action.setChecked(True)
            action.opacity = i / 100.0
            self.connect(action, QtCore.SIGNAL("triggered()"),
                         self.change_opacity)
            self.opacity_group.addAction(action)
            self.menuTransparency.addAction(action)
        # map with menu =======================================================
        self.map.contextmenu = TrayContextMenu(self.trayicon)
        def mapContextMenuEvent(event):
            self.map.contextmenu.exec_(self.mapToGlobal
                                       (QPoint(event.x(),event.y())))
        self.map.contextMenuEvent = mapContextMenuEvent
        self.map.connect(self.map, Qt.SIGNAL("linkClicked(const QUrl&)"), 
                         self.map_link_clicked)
        # end map =============================================================
        self.filewatcher_thread = filewatcher.FileWatcher(self.path_to_logs,
                                                          60*60*24)
        self.connect(self.filewatcher_thread, QtCore.SIGNAL("fchange"),
                     self.logfile_changed)
        self.filewatcher_thread.start()
        self.maptimer = QtCore.QTimer(self)
        self.connect(self.maptimer, QtCore.SIGNAL("timeout()"), self.update_map)
        self.maptimer.start(1000)
        self.connect(trayicon, Qt.SIGNAL("alarm_distance"), 
                     self.change_alarm_distance)
        self.connect(self.actionFrameless_Window, Qt.SIGNAL("triggered()"),
                     self.change_frameless)
        self.connect(trayicon, Qt.SIGNAL("change_frameless"),
                     self.change_frameless)        
        self.connect(self.frameButton, Qt.SIGNAL("clicked()"),
                     self.change_frameless)
        self.frameButton.setVisible(False)
        self.is_frameless = None  # we need this because 2 places to change
        self.alarm_distance = 0
        self.connect(self.actionActivate_Sound, Qt.SIGNAL("triggered()"),
                     self.change_sound)
        if not sound.SOUND_AVAILABLE:
            self.change_sound(disable=True)
        else:
            self.change_sound()
        # load something from cache =====================================
        c = Cache()
        self.known_playernames = c.get_from_cache("known_playernames")
        if self.known_playernames:
            self.known_playernames = set(self.known_playernames.split(","))
        else:
            self.known_playernames = set()
        roomnames = c.get_from_cache("roomnames")
        if roomnames:
            roomnames = roomnames.split(",")
        else:
            roomnames = (u"TheCitadel", u"North Provi Intel")
            c.put_into_cache("roomnames", u",".join(roomnames))
        try:
            settings = c.get_from_cache("settings")
            if settings:
                settings = eval(settings)
                for setting in settings:
                    obj = self if not setting[0] else getattr(self, setting[0])
                    getattr(obj, setting[1])(setting[2])
        except Exception as e:
            self.trayicon.showMessage("Can't remember", 
                 "Something went wrong if I load my last state:\n"\
                 "{0}".format(str(e)), 1)
        # load cache ends ===============================================
        self.connect(self.actionQuit, Qt.SIGNAL("triggered()"), self.close)
        self.connect(self.trayicon, Qt.SIGNAL("quit"), self.close)
        self.chatparser = chatparser.ChatParser(self.path_to_logs,
                          roomnames, self.systems)
        version_check_thread = drachenjaeger.NotifyNewVersionThread()
        version_check_thread.connect(version_check_thread, 
                Qt.SIGNAL("newer_version"), self.notify_newer_version)
        version_check_thread.run()
        
    def notify_newer_version(self, newest_version):
        self.trayicon.showMessage("Newer Version", 
                ("A newer Version of VINTEL is available.\n"
                 "Find the URL in the info!"), 1)
        
    def change_chat_visibility(self, new_value=None):
        if new_value is not None:
            self.action_show_chat.setChecked(new_value)
        self.chatbox.setVisible(self.action_show_chat.isChecked())
        
    def change_opacity(self, value=None):
        if value:
            for action in self.opacity_group.actions():
                if action.opacity == value:
                    action.setChecked(True)
        action = self.opacity_group.checkedAction()
        self.setWindowOpacity(action.opacity)

    def change_sound(self, new_value=None, disable=False):
        if disable:
            self.actionActivate_Sound.setChecked(False)
            self.actionActivate_Sound.setEnabled(False)
            QtGui.QMessageBox.warning(None, "Sound disabled", 
            "I can't find the lib 'pygame' which I use to play sounds, "\
            "so I have to disable the soundsystem.\n"\
            "If you want sound, please install the 'pygame' library.", "OK")
        else:
            if new_value is not None:
                self.actionActivate_Sound.setChecked(new_value)
            sound.sound_active = self.actionActivate_Sound.isChecked()
        
    def add_message_to_intelchat(self, message):
        scroll_to_bottom = False
        if (self.chatListWidget.verticalScrollBar().value() ==
            self.chatListWidget.verticalScrollBar().maximum()):
            scroll_to_bottom = True
        entry = ChatEntry(message)
        listWidgetItem = QtGui.QListWidgetItem(self.chatListWidget)
        listWidgetItem.setSizeHint(entry.sizeHint())
        self.chatListWidget.addItem(listWidgetItem)
        self.chatListWidget.setItemWidget(listWidgetItem, entry)
        self.avatar_find_thread.add_chatentry(entry)
        self.chatentries.append(entry)
        self.connect(entry, Qt.SIGNAL("mark_system"), self.mark_system_on_map)
        self.emit(Qt.SIGNAL("chatmessage_added"), entry)
        if scroll_to_bottom:
            self.chatListWidget.scrollToBottom()

    def change_always_on_top(self, new_value=None):
        self.hide()
        if new_value is not None:
            self.actionAlways_on_top.setChecked(new_value)
        always_on_top = self.actionAlways_on_top.isChecked()
        if always_on_top:
            self.setWindowFlags(self.windowFlags() |
                                QtCore.Qt.WindowStaysOnTopHint)
        else:
            self.setWindowFlags(self.windowFlags() &
                                (~QtCore.Qt.WindowStaysOnTopHint))
        self.show()
        
    def change_frameless(self, new_value=None):
        self.hide()
        if new_value is None:
            if self.is_frameless is None:
                self.is_frameless = False
            new_value = not self.is_frameless
        if new_value:
            self.setWindowFlags(QtCore.Qt.FramelessWindowHint)
            self.menubar.setVisible(False)
            self.frameButton.setVisible(True)
        else:
            self.setWindowFlags(self.windowFlags() &
                                (~QtCore.Qt.FramelessWindowHint))
            self.menubar.setVisible(True)
            self.frameButton.setVisible(False)
        self.change_always_on_top(new_value)
        self.is_frameless = new_value
        self.actionFrameless_Window.setChecked(new_value)
        for cm in TrayContextMenu.INSTANCES:
            cm.frameless_check.setChecked(new_value)
        self.show()
            
    def change_show_avatars(self, new_value=None):
        if new_value is not None:
            self.actionShow_Chat_Avatars.setChecked(new_value)
        show = self.actionShow_Chat_Avatars.isChecked()
        ChatEntry.SHOW_AVATAR = show
        for entry in self.chatentries:
            entry.avatar_label.setVisible(show)

    def chat_smaller(self):
        new_size = ChatEntry.TEXTSIZE - 1
        ChatEntry.TEXTSIZE = new_size
        for entry in self.chatentries:
            entry.change_fontsize(new_size)

    def chat_larger(self):
        new_size = ChatEntry.TEXTSIZE + 1
        ChatEntry.TEXTSIZE = new_size
        for entry in self.chatentries:
            entry.change_fontsize(new_size)
            
    def change_alarm_distance(self, distance):
        self.alarm_distance = distance
        for cm in TrayContextMenu.INSTANCES:
            for action in cm.distance_group.actions():
                if action.alarm_distance == distance:
                    action.setChecked(True)
        self.trayicon.alarm_distance = distance
            
    def change_jumpbridge_view(self):
        self.dotlan.change_jumpbrigde_visibility()
        self.update_map()
            
    def clipboard_changed(self, mode):
        if mode == 0 and self.action_kos_clipboard_active.isChecked():
            content = unicode(self.clipboard.text())
            last_modified, old_content = self.old_clipboard_content
            if content == old_content and time.time() - last_modified < 3:
                parts = content.split("\n")
                for part in parts:
                    if part in self.known_playernames:
                        check_result = koschecker.check(parts)
                        text = koschecker.result_to_text(check_result, True)
                        if len(text) > 0:
                            sound.play_sound("beep")
                        else:
                            text = "Noone KOS"
                        self.trayicon.showMessage("Your KOS-Check", text, 1)
                        self.old_clipboard_content = (0, "")
                        break
            else:
                self.old_clipboard_content = (time.time(), content)
                
    def closeEvent(self, event):
        """ writing the cache before closing the window """
        c = Cache()
        # known playernames
        if self.known_playernames:
            value = ",".join(self.known_playernames)
            c.put_into_cache("known_playernames", value, 60*60*24*365)
        # program state to cache (to read it on next startup)
        settings = (
            (None, "restoreGeometry", str(self.saveGeometry())),
            (None, "restoreState", str(self.saveState())),
            (None, "change_opacity",
             self.opacity_group.checkedAction().opacity),
            (None, "change_always_on_top",
             self.actionAlways_on_top.isChecked()),
            ("splitter", "restoreGeometry",
             str(self.splitter.saveGeometry())),
            ("splitter", "restoreState",
             str(self.splitter.saveState())),
            (None, "change_show_avatars",
             self.actionShow_Chat_Avatars.isChecked()),
            (None, "change_alarm_distance", self.alarm_distance),
            ("action_kos_clipboard_active", "setChecked", 
             self.action_kos_clipboard_active.isChecked()),
            (None, "change_sound", self.actionActivate_Sound.isChecked()),
            (None, "change_chat_visibility", self.action_show_chat.isChecked()),
            ("map", "setZoomFactor", self.map.zoomFactor()),
            (None, "set_init_map_scrollposition", 
             (self.map.page().mainFrame().scrollPosition().x(),
              self.map.page().mainFrame().scrollPosition().y())),
            (None, "change_frameless", self.actionFrameless_Window.isChecked()),
        )
        c.put_into_cache("settings", str(settings),
                         60*60*24*365)
        event.accept()
          
    def map_link_clicked(self, url):
        systemname = unicode(url.path().split("/")[-1]).upper()
        system = self.systems[str(systemname)]
        sc = SystemChat(self, SystemChat.SYSTEM, system, self.chatentries,
                        self.known_playernames)
        sc.connect(self, Qt.SIGNAL("chatmessage_added"), sc.add_chatentry)
        sc.connect(self, Qt.SIGNAL("avatar_loaded"), sc.new_avatar_available)
        sc.connect(sc, Qt.SIGNAL("location_set"), self.set_location)
        sc.show()
        
    def mark_system_on_map(self, systemname):
        self.systems[unicode(systemname)].mark()
        self.update_map()

    def set_location(self, char, new_system):
        for system in self.systems.values():
            system.remove_located_character(char)
        if not new_system == "?" and new_system in self.systems:
            self.systems[new_system].add_located_character(char)
            self.set_map_content(self.dotlan.svg)
        
    def set_map_content(self, content):
        if self.init_map_position is None:
            scrollposition = self.map.page().mainFrame().scrollPosition()
        else:
            scrollposition = self.init_map_position
            self.init_map_position = None
        self.map.setContent(content)
        self.map.page().mainFrame().setScrollPosition(scrollposition)
        self.map.page().setLinkDelegationPolicy(QWebPage.DelegateAllLinks)
        
    def set_init_map_scrollposition(self, xy):
        self.init_map_position = QPoint(xy[0], xy[1])
        
    def show_chatroom_chooser(self):
        chooser = ChatroomsChooser(self)
        chooser.connect(chooser, Qt.SIGNAL("rooms_changed"),
                        self.changed_roomnames)
        chooser.show()

    def changed_roomnames(self, new_roomnames):
        cache = Cache()
        cache.put_into_cache("roomnames", u",".join(new_roomnames))
        self.chatparser.rooms = new_roomnames
        
    def show_info(self):
        info_dialog = QtGui.QDialog(self)
        uic.loadUi(resource_path("vi/ui/Info.ui"), info_dialog)
        info_dialog.version_label.setText(
            u"Version: {0}".format(VERSION))
        info_dialog.logo_label.setPixmap(QtGui.QPixmap(
            resource_path("vi/ui/res/logo.png")))
        info_dialog.connect(info_dialog.close_button, Qt.SIGNAL("clicked()"), 
                     info_dialog.accept)
        info_dialog.show()

    def systray_activated(self, reason):
        if reason == QtGui.QSystemTrayIcon.Trigger:
            if self.isMinimized():
                self.showNormal()
                self.activateWindow()
            elif not self.isActiveWindow():
                self.activateWindow()
            else:
                self.showMinimized()

    def update_avatar_on_chatentry(self, chatentry, avatar_data):
        updated = chatentry.update_avatar(avatar_data)
        if not updated:
            self.avatar_find_thread.add_chatentry(chatentry, clear_cache=True)
        else:
            self.emit(Qt.SIGNAL("avatar_loaded"), chatentry.message.user,
                      avatar_data)
        
    def update_map(self):
        self.set_map_content(self.dotlan.svg)

    def zoomMapIn(self):
        self.map.setZoomFactor(self.map.zoomFactor() + 0.1)

    def zoomMapOut(self):
        self.map.setZoomFactor(self.map.zoomFactor() - 0.1)


    def logfile_changed(self, path):
        messages = self.chatparser.file_modified(path)
        for message in messages:
            # if players location changed
            if message.status == message.LOCATION:
                self.known_playernames.add(message.user)
                self.set_location(message.user, message.systems[0])
            elif message.status == message.SOUNDTEST \
              and message.user in self.known_playernames:
                words = message.message.split()
                if len(words) > 1:
                    sound.play_sound(words[1])
            elif message.status == message.KOS_STATUS_REQUEST:
                text = message.message.replace("xxx", "").replace("XXX", "")
                text = text.replace("  ", ",")
                parts = (name.strip() for name in text.split(","))
                check_result = koschecker.check(parts)
                text = koschecker.result_to_text(check_result, False)
                if text.startswith("KOS:"):
                    sound.play_sound("beep")
                self.trayicon.showMessage("A xxx KOS-Check", text, 1)
            # if it is a 'normal' chat message
            elif (message.user != "EVE-System"
                and message.status != message.IGNORE):
                self.add_message_to_intelchat(message)
                if message.systems:
                    for system in message.systems:
                        systemname = system.name
                        self.dotlan.systems[systemname].set_status(
                            message.status)
                        if message.status in (message.REQUEST, message.ALARM)\
                            and message.user not in self.known_playernames:
                            alarm_distance = self.alarm_distance if \
                                message.status == message.ALARM else 0
                            for nsystem, data in system.get_neighbours(
                                alarm_distance).items():
                                distance = data["distance"]
                                chars = nsystem.get_located_characters()
                                if len(chars) > 0 and message.user not in chars:
                                    self.trayicon.show_notification(
                                        message, system.name, ", ".join(chars),
                                        distance)
                self.set_map_content(self.dotlan.svg)


class ChatroomsChooser(QtGui.QDialog):
    
    def __init__(self, parent):
        QtGui.QDialog.__init__(self, parent)
        uic.loadUi(resource_path("vi/ui/ChatroomsChooser.ui"), self)
        self.connect(self.default_button, Qt.SIGNAL("clicked()"), 
                     self.set_defaults)
        self.connect(self.cancel_button, Qt.SIGNAL("clicked()"), self.accept)
        self.connect(self.save_button, Qt.SIGNAL("clicked()"), 
                     self.save_clicked)
        cache = Cache()
        roomnames = cache.get_from_cache("roomnames")
        if not roomnames:
            roomnames = u"TheCitadel, North Provi Intel"
        self.roomnames_field.setPlainText(roomnames)
        
    def save_clicked(self):
        text = unicode(self.roomnames_field.toPlainText())
        rooms = [unicode(name.strip()) for name in text.split(",")]
        self.emit(Qt.SIGNAL("rooms_changed"), rooms)
        self.accept()
        
    def set_defaults(self):
        self.roomnames_field.setPlainText(u"TheCitadel,North Provi Intel")
        
        
class SystemChat(QtGui.QDialog):

    SYSTEM = 0
    
    def __init__(self, parent, chat_type, selector, chatentries,
                 known_playernames):
        QtGui.QDialog.__init__(self, parent)
        uic.loadUi(resource_path("vi/ui/SystemChat.ui"), self)
        self.parent = parent
        self.chat_type = 0
        self.selector = selector
        self.chatentries = []
        for entry in chatentries:
            self.add_chatentry(entry)
        titlename = ""
        if self.chat_type == SystemChat.SYSTEM:
            titlename = self.selector.name
            self.system = selector
        for name in known_playernames:
            self.playernames_box.addItem(name)
        self.setWindowTitle("Chat for {0}".format(titlename))
        self.connect(self.close_button, Qt.SIGNAL("clicked()"), 
                     self.close_dialog)
        self.connect(self.alarm_button, Qt.SIGNAL("clicked()"), 
                     self.set_system_alarm)
        self.connect(self.clear_button, Qt.SIGNAL("clicked()"), 
                     self.set_system_clear)
        self.connect(self.location_button, Qt.SIGNAL("clicked()"),
                     self.location_set)

    def _add_message_to_chat(self, message, avatar_pixmap):
        scroll_to_bottom = False
        if (self.chat.verticalScrollBar().value() ==
            self.chat.verticalScrollBar().maximum()):
            scroll_to_bottom = True
        entry = ChatEntry(message)
        entry.avatar_label.setPixmap(avatar_pixmap)
        listWidgetItem = QtGui.QListWidgetItem(self.chat)
        listWidgetItem.setSizeHint(entry.sizeHint())
        self.chat.addItem(listWidgetItem)
        self.chat.setItemWidget(listWidgetItem, entry)
        self.chatentries.append(entry)
        self.connect(entry, Qt.SIGNAL("mark_system"),
                     self.parent.mark_system_on_map)
        if scroll_to_bottom:
            self.chat.scrollToBottom()

    def add_chatentry(self, entry):
        if self.chat_type == SystemChat.SYSTEM:
            message = entry.message
            avatar_pixmap = entry.avatar_label.pixmap()
            if self.selector in message.systems:
                self._add_message_to_chat(message, avatar_pixmap)

    def location_set(self):
        char = unicode(self.playernames_box.currentText())
        self.emit(Qt.SIGNAL("location_set"), char, self.system.name)

    def new_avatar_available(self, charname, avatar_data):
        for entry in self.chatentries:
            if entry.message.user == charname:
                entry.update_avatar(avatar_data)
        
    def set_system_alarm(self):
        self.system.set_status(self.system.ALARM)
        self.parent.update_map()
            
    def set_system_clear(self):
        self.system.set_status(self.system.CLEAR)
        self.parent.update_map()
          
    def close_dialog(self):
        self.accept()


class ChatEntry(QtGui.QWidget):

    TEXTSIZE = 9
    SHOW_AVATAR = True

    def __init__(self, message):
        QtGui.QWidget.__init__(self)
        uic.loadUi(resource_path("vi/ui/ChatEntry.ui"), self)
        self.avatar_label.setPixmap(QtGui.QPixmap(
            resource_path("vi/ui/res/qmark.png")))
        self.message = message
        self.update_text()
        self.connect(self.text_label, QtCore.SIGNAL("linkActivated(QString)"), 
                     self.link_clicked)
        self.change_fontsize(self.TEXTSIZE)
        if not ChatEntry.SHOW_AVATAR:
            self.avatar_label.setVisible(False)
    
    def link_clicked(self, link):
        link = unicode(link)
        function, parameter = link.split("/", 1)
        if function == "mark_system":
            self.emit(QtCore.SIGNAL("mark_system"), parameter)
        if function == "link":
            webbrowser.open(parameter)

    def update_text(self):
        time = datetime.datetime.strftime(self.message.timestamp, "%H:%M:%S")
        text = u"<small>{time} - <b>{user}</b> - <i>{room}</i></small><br>{text}".format(
            user=self.message.user, room=self.message.room, time=time,
            text=self.message.message)
        self.text_label.setText(text)

    def update_avatar(self, avatar_data):
        image = QImage.fromData(avatar_data)
        pixmap = QPixmap.fromImage(image)
        if pixmap.isNull():
            return False
        scaled_avatar = pixmap.scaled(32, 32)
        self.avatar_label.setPixmap(scaled_avatar)
        return True

    def change_fontsize(self, new_size):
        font = self.text_label.font()
        font.setPointSize(new_size)
        self.text_label.setFont(font)


